from __future__ import annotations

import pytest
import sqlalchemy as sa
from flask import Flask

from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy.session import Session


def test_scope(app: Flask, db: SQLAlchemy) -> None:
    with pytest.raises(RuntimeError):
        db.session()

    with app.app_context():
        first = db.session()
        second = db.session()
        assert first is second
        assert isinstance(first, Session)

    with app.app_context():
        third = db.session()
        assert first is not third


def test_custom_scope(app: Flask) -> None:
    count = 0

    def scope() -> int:
        nonlocal count
        count += 1
        return count

    db = SQLAlchemy(app, session_options={"scopefunc": scope})

    with app.app_context():
        first = db.session()
        second = db.session()
        assert first is not second  # a new scope is generated on each call
        first.close()
        second.close()


@pytest.mark.usefixtures("app_ctx")
def test_session_class(app: Flask) -> None:
    class CustomSession(Session):
        pass

    db = SQLAlchemy(app, session_options={"class_": CustomSession})
    assert isinstance(db.session(), CustomSession)


@pytest.mark.usefixtures("app_ctx")
def test_session_uses_bind_key(app: Flask) -> None:
    app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
    db = SQLAlchemy(app)

    class User(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    class Post(db.Model):
        __bind_key__ = "a"
        id = sa.Column(sa.Integer, primary_key=True)

    assert db.session.get_bind(mapper=User) is db.engine
    assert db.session.get_bind(mapper=Post) is db.engines["a"]


@pytest.mark.usefixtures("app_ctx")
def test_get_bind_inheritance(app: Flask) -> None:
    app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
    db = SQLAlchemy(app)

    class User(db.Model):
        __bind_key__ = "a"
        id = sa.Column(sa.Integer, primary_key=True)
        type = sa.Column(sa.String, nullable=False)

        __mapper_args__ = {"polymorphic_on": type, "polymorphic_identity": "user"}

    class Admin(User):
        id = sa.Column(sa.ForeignKey(User.id), primary_key=True)
        org = sa.Column(sa.String, nullable=False)

        __mapper_args__ = {"polymorphic_identity": "admin"}

    db.create_all()
    db.session.add(Admin(org="pallets"))
    db.session.commit()
    admin = db.session.execute(db.select(Admin)).scalar_one()
    db.session.expire(admin)
    assert admin.org == "pallets"



from __future__ import annotations

import typing as t
from pathlib import Path

import pytest
import sqlalchemy as sa
from flask import Flask
from flask.ctx import AppContext

from flask_sqlalchemy import SQLAlchemy


@pytest.fixture
def app(request: pytest.FixtureRequest, tmp_path: Path) -> Flask:
    app = Flask(request.module.__name__, instance_path=str(tmp_path / "instance"))
    app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite://"
    app.config["SQLALCHEMY_RECORD_QUERIES"] = False
    return app


@pytest.fixture
def app_ctx(app: Flask) -> t.Generator[AppContext, None, None]:
    with app.app_context() as ctx:
        yield ctx


@pytest.fixture
def db(app: Flask) -> SQLAlchemy:
    return SQLAlchemy(app)


@pytest.fixture
def Todo(app: Flask, db: SQLAlchemy) -> t.Any:
    class Todo(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)
        title = sa.Column(sa.String)

    with app.app_context():
        db.create_all()

    yield Todo

    with app.app_context():
        db.drop_all()



from __future__ import annotations

import pytest
import sqlalchemy as sa
from flask import Flask

from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy.record_queries import get_recorded_queries


@pytest.mark.usefixtures("app_ctx")
def test_query_info(app: Flask) -> None:
    app.config["SQLALCHEMY_RECORD_QUERIES"] = True
    db = SQLAlchemy(app)

    class Example(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    db.create_all()
    db.session.execute(sa.select(Example).filter(Example.id < 5)).scalars()
    info = get_recorded_queries()[-1]
    assert info.statement is not None
    assert "SELECT" in info.statement
    assert "FROM example" in info.statement
    assert info.parameters[0][0] == 5
    assert info.duration == info.end_time - info.start_time
    assert "tests/test_record_queries.py:" in info.location
    assert "(test_query_info)" in info.location



from __future__ import annotations

import typing as t

import pytest
import sqlalchemy as sa
import sqlalchemy.orm
from flask import Flask

from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy.model import DefaultMeta
from flask_sqlalchemy.model import Model


def test_default_model_class(db: SQLAlchemy) -> None:
    assert db.Model.query_class is db.Query
    assert db.Model.metadata is db.metadata
    assert issubclass(db.Model, Model)
    assert isinstance(db.Model, DefaultMeta)


def test_custom_model_class(app: Flask) -> None:
    class CustomModel(Model):
        pass

    db = SQLAlchemy(app, model_class=CustomModel)
    assert issubclass(db.Model, CustomModel)
    assert isinstance(db.Model, DefaultMeta)


@pytest.mark.usefixtures("app_ctx")
@pytest.mark.parametrize("base", [Model, object])
def test_custom_declarative_class(app: Flask, base: t.Any) -> None:
    class CustomMeta(DefaultMeta):
        pass

    CustomModel = sa.orm.declarative_base(cls=base, name="Model", metaclass=CustomMeta)
    db = SQLAlchemy(app, model_class=CustomModel)
    assert db.Model is CustomModel
    assert db.Model.query_class is db.Query
    assert "query" in db.Model.__dict__


@pytest.mark.usefixtures("app_ctx")
def test_model_repr(db: SQLAlchemy) -> None:
    class User(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    db.create_all()
    user = User()
    assert repr(user) == f"<User (transient {id(user)})>"
    db.session.add(user)
    assert repr(user) == f"<User (pending {id(user)})>"
    db.session.flush()
    assert repr(user) == f"<User {user.id}>"



from __future__ import annotations

import typing as t
import warnings

import pytest
import sqlalchemy as sa
import sqlalchemy.exc
from flask import Flask
from werkzeug.exceptions import NotFound

from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy.query import Query


@pytest.fixture(autouse=True)
def ignore_query_warning() -> t.Generator[None, None, None]:
    if hasattr(sa.exc, "LegacyAPIWarning"):
        with warnings.catch_warnings():
            exc = sa.exc.LegacyAPIWarning
            warnings.simplefilter("ignore", exc)
            yield
    else:
        yield


@pytest.mark.usefixtures("app_ctx")
def test_get_or_404(db: SQLAlchemy, Todo: t.Any) -> None:
    item = Todo()
    db.session.add(item)
    db.session.commit()
    assert Todo.query.get_or_404(1) is item

    with pytest.raises(NotFound):
        Todo.query.get_or_404(2)


@pytest.mark.usefixtures("app_ctx")
def test_first_or_404(db: SQLAlchemy, Todo: t.Any) -> None:
    db.session.add(Todo(title="a"))
    db.session.commit()
    assert Todo.query.filter_by(title="a").first_or_404().title == "a"

    with pytest.raises(NotFound):
        Todo.query.filter_by(title="b").first_or_404()


@pytest.mark.usefixtures("app_ctx")
def test_one_or_404(db: SQLAlchemy, Todo: t.Any) -> None:
    db.session.add(Todo(title="a"))
    db.session.add(Todo(title="b"))
    db.session.add(Todo(title="b"))
    db.session.commit()
    assert Todo.query.filter_by(title="a").one_or_404().title == "a"

    with pytest.raises(NotFound):
        # MultipleResultsFound
        Todo.query.filter_by(title="b").one_or_404()

    with pytest.raises(NotFound):
        # NoResultFound
        Todo.query.filter_by(title="c").one_or_404()


@pytest.mark.usefixtures("app_ctx")
def test_paginate(db: SQLAlchemy, Todo: t.Any) -> None:
    db.session.add_all(Todo() for _ in range(150))
    db.session.commit()
    p = Todo.query.paginate()
    assert p.total == 150
    assert len(p.items) == 20
    p2 = p.next()
    assert p2.page == 2
    assert p2.total == 150


@pytest.mark.usefixtures("app_ctx")
def test_default_query_class(db: SQLAlchemy) -> None:
    class Parent(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)
        children1 = db.relationship("Child", backref="parent1", lazy="dynamic")

    class Child(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)
        parent_id = sa.Column(sa.ForeignKey(Parent.id))  # type: ignore[var-annotated]
        parent2 = db.relationship(
            Parent,
            backref=db.backref("children2", lazy="dynamic", viewonly=True),
            viewonly=True,
        )

    p = Parent()
    assert type(Parent.query) is Query
    assert isinstance(p.children1, Query)
    assert isinstance(p.children2, Query)
    assert isinstance(db.session.query(Child), Query)


@pytest.mark.usefixtures("app_ctx")
def test_custom_query_class(app: Flask) -> None:
    class CustomQuery(Query):
        pass

    db = SQLAlchemy(app, query_class=CustomQuery)

    class Parent(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)
        children1 = db.relationship("Child", backref="parent1", lazy="dynamic")

    class Child(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)
        parent_id = sa.Column(sa.ForeignKey(Parent.id))  # type: ignore[var-annotated]
        parent2 = db.relationship(
            Parent,
            backref=db.backref("children2", lazy="dynamic", viewonly=True),
            viewonly=True,
        )

    p = Parent()
    assert type(Parent.query) is CustomQuery
    assert isinstance(p.children1, CustomQuery)
    assert isinstance(p.children2, CustomQuery)
    assert isinstance(db.session.query(Child), CustomQuery)



from __future__ import annotations

import typing as t

import pytest

from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy.cli import add_models_to_shell


@pytest.mark.usefixtures("app_ctx")
def test_shell_context(db: SQLAlchemy, Todo: t.Any) -> None:
    context = add_models_to_shell()
    assert context["db"] is db
    assert context["Todo"] is Todo



from __future__ import annotations

import typing as t

import pytest
import sqlalchemy as sa
from flask import Flask

from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy.track_modifications import before_models_committed
from flask_sqlalchemy.track_modifications import models_committed

pytest.importorskip("blinker")


@pytest.mark.usefixtures("app_ctx")
def test_track_modifications(app: Flask) -> None:
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = True
    db = SQLAlchemy(app)

    class Example(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)
        data = sa.Column(sa.String)

    db.create_all()
    before: list[tuple[t.Any, str]] = []
    after: list[tuple[t.Any, str]] = []

    def before_commit(sender: Flask, changes: list[tuple[t.Any, str]]) -> None:
        nonlocal before
        before = changes

    def after_commit(sender: Flask, changes: list[tuple[t.Any, str]]) -> None:
        nonlocal after
        after = changes

    connect_before = before_models_committed.connected_to(before_commit, app)
    connect_after = models_committed.connected_to(after_commit, app)

    with connect_before, connect_after:
        item = Example()

        db.session.add(item)
        assert not before
        assert not after

        db.session.commit()
        assert len(before) == 1
        assert before[0] == (item, "insert")
        assert before == after

        db.session.remove()
        item = db.session.get(Example, 1)  # type: ignore[assignment]
        item.data = "test"  # type: ignore[assignment]
        db.session.commit()
        assert len(before) == 1
        assert before[0] == (item, "update")
        assert before == after

        db.session.remove()
        item = db.session.get(Example, 1)  # type: ignore[assignment]
        db.session.delete(item)
        db.session.commit()
        assert len(before) == 1
        assert before[0] == (item, "delete")
        assert before == after
